/*	$Id: walkMainCodels.cc 2009/10/12 18:08:21 mallet $ */

/*
 * Copyright (c) 2007 Mathieu Poirier                 --  Thu Apr  5 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
static const char id[] = "$Id: walkMainCodels.cc 2009/10/12 18:08:21 mallet $";

#include <portLib.h>

#include <err.h>
#include <unistd.h>
#include <stddef.h>
#include <limits.h>
#include <sys/utsname.h>
#include <iostream>

#include "server/walkHeader.h"

#include "KineoModel/kppLicense.h"
#include "KineoKCDModel/kppKCDPolyhedron.h"
#include "KineoModel/kppJointComponent.h"
#include "KineoWorks2/kwsConfig.h"
#include "KineoWorks2/kwsPickerBasic.h"
#include "KineoWorks2/kwsShooterConfigSpace.h"

#include "KineoWorks2/kwsShooterPath.h"

#include "kwsPlus/hppVisRdmBuilder.h"
#include "kwsPlus/hppShooterActiveDof.h"

#include "hppWalkPlanner/hppWalkPlanner.h"

#include "hppWalkFootPlanner/hppWalkSteeringMethod.h"
#include "hppWalkFootPlanner/hppWalkDistance.h"

#include "hppModel/hppDevice.h"
#include "hppCorbaServer/hppciOpenHrp.h"
#include "kwsPlus/flicSteeringMethod.h"
#include "kwsPlus/flicDistance.h"

#include "hrp2Struct.h"
#include "netshring.h"

#include "hrp2_14/hrp2_14.h"

// Select level of verbosity at configuration.
#if DEBUG==2
#define ODEBUG2(x) std::cout << "walkMainCodels:" << x << std::endl
#define ODEBUG1(x) std::cerr << "walkMainCodels:" << x << std::endl
#elif DEBUG==1
#define ODEBUG2(x)
#define ODEBUG1(x) std::cerr << "walkMainCodels:" << x << std::endl
#else
#define ODEBUG2(x)
#define ODEBUG1(x)
#endif


/* --- local data ---------------------------------------------------- */

#define TOR_OBSTACLE	(0.015)
#define PENETRATION_BOX	(0.05)
#define OPT_MAX_NB	(40)

static netshring_id_t shtarget;		/**< shared trajectory buffer */
static ChppWalkPlanner *hppPlanner;	/**< hpp planner */


/* --- setDefaultRoadmapBuilderAndOptimizer -------------------------- */

static ktStatus
setDefaultRoadmapBuilderAndOptimizer()
{
  // Set default tools for the robot bouding box.
  // Create steering method
  double distanceThreshold = 5.0;
  ChppWalkSteeringMethodShPtr steeringMethod = ChppWalkSteeringMethod::create ( distanceThreshold );
  if (!steeringMethod)
    return KD_ERROR;
  hppPlanner->steeringMethodIthProblem(0, steeringMethod);

  const ChppDeviceShPtr robotBox =
    KIT_DYNAMIC_PTR_CAST(ChppDevice, hppPlanner->robotIthProblem(0));

  if (!robotBox) {
    ODEBUG1(":walkInitMain: failed to get robot box.");
    return KD_ERROR;
  }

  //create RoadMap
  CkwsRoadmapShPtr kwsPlusRoadmap = CkwsRoadmap::create(robotBox) ;
  if (!kwsPlusRoadmap) {
    ODEBUG1(":walkInitMain: failed to create roadmap.");
    return KD_ERROR;
  }

  //create distance evaluator
  ChppWalkDistanceShPtr distanceFunction = ChppWalkDistance::create ( distanceThreshold );
  if (!distanceFunction)
  {
    ODEBUG1(":walkInitMain: failed to create the distance evaluator");
    return KD_ERROR;
  }

  // Give distance function to robotBox
  robotBox->distance(distanceFunction);

  //create Diffusion Node Picker
  CkwsDiffusionNodePickerShPtr picker = CkwsPickerBasic::create ();

  //create Diffusion Shooter
  //CkwsShooterConfigSpaceShPtr shooterCfgSp =  CkwsShooterConfigSpace::create();
  ChppShooterActiveDofShPtr shooterCfgSp=ChppShooterActiveDof::create();
  if ( !shooterCfgSp )
  {
    ODEBUG1(":walkInitMain: failed to create the shooterCfgSp");
    return KD_ERROR;
  }

  // Create a roadmap builder strategie
   ChppVisRdmBuilderShPtr kwsRoadmapBuilder = ChppVisRdmBuilder::create(kwsPlusRoadmap, PENETRATION_BOX, shooterCfgSp);

  if (!kwsRoadmapBuilder) {
    ODEBUG1(":walkInitMain: failed to create the roadmap.");
    return KD_ERROR;
  }

  kwsRoadmapBuilder->distance(distanceFunction);


  if (CkwsValidatorDPCollision::setPenetration(kwsRoadmapBuilder->builderDirectPathValidator(),PENETRATION_BOX) != KD_OK) {
    ODEBUG1(":walkInitMain: failed to set the penetration.");
    return KD_ERROR;
  }

  //kwsRoadmapBuilder->diffuseFromProblemStart (true);
  //kwsRoadmapBuilder->diffuseFromProblemGoal (true);

  hppPlanner->roadmapBuilderIthProblem(0, kwsRoadmapBuilder);

  // Create a Path Optimizer
  CkwsLoopOptimizerShPtr kwsOptimizer = CkwsRandomOptimizer::create();
  kwsOptimizer->distance(distanceFunction);
  kwsOptimizer->shortcutMethod(CkwsShortcutDirect::create());
  kwsOptimizer->maxNbLoop(OPT_MAX_NB);

  //kwsOptimizer->doesClear(true);

  //CkwsAdaptiveShortcutOptimizerShPtr kwsOptimizer =  CkwsAdaptiveShortcutOptimizer::create();
  //kwsOptimizer->distance(distanceFunction);
  //kwsOptimizer->shortcutMethod(CkwsShortcutDirect::create());

  ODEBUG2(" max nb loop for the optimizer "<< kwsOptimizer->maxNbLoop ());

  hppPlanner->pathOptimizerIthProblem(0, kwsOptimizer) ;


  return KD_OK;
}


/* --- walkMainInit -------------------------------------------------- */

/** Initialization codel (fIDS, ...)
 *
 * \returns OK or ERROR
 */

STATUS
walkMainInit(int *report)
{
  SDI_F->initialized = 0;

  /* create shared trajectory buffer */
  shtarget = netshring_create(WALK_TARGET_CONFIG_NAME);
  if (!shtarget) {
    warnx("cannot create shared buffer");
    return ERROR;
  }

  return OK;
}


/* --- walkMainEnd --------------------------------------------------- */

/** Termination codel
 *
 * \returns OK or ERROR
 */

STATUS
walkMainEnd(void)
{
  /* delete shared trajectory buffer */
  if (shtarget) {
    netshring_delete(shtarget);
    shtarget = NULL;
  }

  return OK;
}


/* --- Init ---------------------------------------------------------- */

/** walkInitMain  -  codel EXEC of Init
 *
 * Reports:      OK
 *              S_walk_NAMESERVER_NOT_FOUND
 *              S_walk_OPENHRP_NOT_FOUND
 *              S_walk_CORBA_NOT_FOUND
 *              S_walk_KINEOWORKS_LICENSE_NOT_FOUND
 *              S_walk_MODELLOADERJRL_NOT_FOUND
 *              S_walk_ALREADY_INITIALIZED
 *              S_walk_COULD_NOT_BUILD_HRP2
 *              S_walk_COULD_NOT_BUILD_OBSTACLE
 *
 * \returns EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
walkInitMain(WalkInitData *init_data, int *report)
{
  // ----------- 1. init KineoWorks  -------------------------------

  /* Kineo insists on looking for the license in the cwd, so make it
   * happy if possible. Don't fail on error so that a user can still cd
   * by hand to the proper directory if it cannot be automatically
   * found. */
#ifdef WALK_GENOM_LICENSEPREFIX
  {
    struct utsname hname;
      char ldir[PATH_MAX];
    char *dot;

    if (!uname(&hname)) do {
      snprintf(ldir, sizeof(ldir)-1,
	       WALK_GENOM_LICENSEPREFIX "/%s", hname.nodename);
      if (!chdir(ldir)) {
	warnx("set cwd to %s", ldir);
	break;
      }

      /* strip dots from fqdn */
      dot = strrchr(hname.nodename, '.');
      if (dot) *dot = '\0';
    } while (dot);
  }
#endif

  if(!CkppLicense::initialize()) {
    cerr << "License for KPP SDK NOT FOUND" << endl;
    *report = S_walk_KINEOWORKS_LICENSE_NOT_FOUND;
    return ETHER;
  }
  cerr << "License for KPP SDK found" << endl;

  printf("yyyy %s\n", init_data->model.name);

  // walk planner
  hppPlanner = new ChppWalkPlanner(WALKGEN_STEP_LENGTH,
				   WALKGEN_STEP_SPACING,
				   WALKGEN_STEP_SECURITY_DISTANCE,
				   WALKGEN_BOX_LATERAL_TOLERANCE,
				   init_data->model.name);
  if (hppPlanner == NULL ) {
    cerr << "COULD NOT CREATE a hppWalkPlanner" << endl;
    *report = S_walk_HPP_PLANNER_NOT_FOUND;
    return ETHER;
  }

  // ---------- 2. Load Robot Model from ModelLoader ------------------

  ChppciOpenHrpClient openHrpCorbaClient(hppPlanner);
  printf("xxxxx %s\n", init_data->model.name);
  if (openHrpCorbaClient.loadHrp2Model(0.02, init_data->model.name) != KD_OK) {
    *report = S_walk_COULD_NOT_BUILD_ROBOT;
    return ETHER;
  }


  // ---------- 3. Initialize the walkPlanner with the devices -------

  CkppDeviceComponentShPtr hrp2 = hppPlanner->robotIthProblem(0);
  CkwsConfigShPtr halfSittingConfig = CkwsConfig::create(hrp2) ;
  double dInitPos[46] = HALFSITTINGPOSITION_RAD_KINEO ;
  std::vector<double>  halfSittingVector (dInitPos  , dInitPos + sizeof(dInitPos) / sizeof(double) );
  halfSittingConfig->setDofValues( halfSittingVector ) ;
  hrp2->setCurrentConfig(*halfSittingConfig);

  // Create bounding box robot and place it at beginning of problem vector.
  hppPlanner->initializeProblem();

  // Initialize x-y bounds of the bounding box of the robot
  CkppDeviceComponentShPtr boundingBox = hppPlanner->robotIthProblem(0);
  if (!boundingBox) {
    cerr << "walk::Init: there is no robot in hppPlanner." << endl;
    *report = S_walk_COULD_NOT_BUILD_ROBOT;
    return ETHER;
  }

  CkppJointComponentShPtr kppJoint = boundingBox->rootJointComponent();
  if (!kppJoint) {
    cerr << "walk::Init: bounding box robot has no root joint." << endl;
    *report = S_walk_COULD_NOT_BUILD_ROBOT;
    return ETHER;
  }

  CkwsJointShPtr kwsJoint = kppJoint->kwsJoint();
  if (kwsJoint->countDofs() != 6) {
    cerr << "walk::Init: Root joint of bounding box is not freeflyer." << endl;
    *report = S_walk_COULD_NOT_BUILD_ROBOT;
    return ETHER;
  }

  kwsJoint->dof(0)->isBounded(true);
  kwsJoint->dof(0)->bounds(init_data->bounds.xMin, init_data->bounds.xMax);
  kwsJoint->dof(1)->isBounded(true);
  kwsJoint->dof(1)->bounds(init_data->bounds.yMin, init_data->bounds.yMax);

  boundingBox->extraDof(0)->CkwsDof::bounds(-2, 2);

  if (setDefaultRoadmapBuilderAndOptimizer() != KD_OK) {
    *report = S_walk_COULD_NOT_SETUP_PROBLEM;
    return ETHER;
  }

  std::vector<CkcdObjectShPtr> listOfObstacle;
  hppPlanner->obstacleList ( listOfObstacle );

  SDI_F->initialized = 1;
  return ETHER;
}

/* --- GoToPosition -------------------------------------------------- */

/** walkLoadObstacleStart  -  codel START of LoadObstacle
 *
 * \return ETHER
 */

ACTIVITY_EVENT
walkLoadObstacleStart(WalkName* obstacleName, int* report)
{

  CkppKCDPolyhedronShPtr polyhedron;
  ChppciOpenHrpClient openHrpCorbaClient(hppPlanner);

  if (openHrpCorbaClient.loadObstacleModel(std::string(obstacleName->name),
					   "obstacle",
					   polyhedron) != KD_OK) {
    *report = S_walk_FAILED_TO_LOAD_OBSTACLE;
    return ETHER;
  }
  hppPlanner->addObstacle(polyhedron);

  return ETHER;
}

/* --- GoToPosition -------------------------------------------------- */

/** walkGoToPositionMain  -  codel EXEC of GoToPosition
 *
 * \return OK
 */

ACTIVITY_EVENT
walkGoToPositionMain(WalkFromToPosition *fromTo, int *report)
{
  if (!SDI_F->initialized) { *report = S_walk_NOT_INITIALIZED; return ETHER; }

  double fromRzRad = fromTo->fromPosition.RzPos * M_PI/180.;
  double toRzRad = fromTo->toPosition.RzPos * M_PI/180.;

  hppPlanner->initConfig(
    fromTo->fromPosition.xPos, fromTo->fromPosition.yPos,fromRzRad);
  hppPlanner->goalConfig(
    fromTo->toPosition.xPos, fromTo->toPosition.yPos, toRzRad);

  if (hppPlanner->solve() != KD_OK) {
    *report = S_walk_NO_SOLUTION;
    return ETHER;
  }

  vector<CkwsConfigShPtr> path;
  vector<double> vecConfig;
  vector<double> walkTime;
  vector< vector<double> > zmp;
  vector< vector<double> > waist;
  unsigned int walkIndex, i, nbConfig;

  i=0;
  nbConfig = 0;

  STATUS s = hppPlanner->solution(path, walkTime, zmp, waist);

  for (walkIndex = 0; walkIndex < walkTime.size(); walkIndex=walkIndex+20) {
    path[walkIndex]->getDofValues ( vecConfig );
    nbConfig++;

    for (int j=0; j < (int) vecConfig.size(); j++){
      SDI_F->walkHrp2Traj.dof[i]=vecConfig[j];
      i++;
    }
  }
  SDI_F->walkHrp2Traj.nbDof=vecConfig.size();
  SDI_F->walkHrp2Traj.nbConfig=nbConfig;
  ODEBUG2("walkHrp2Traj.nbDof = " << vecConfig.size());
  ODEBUG2("walkHrp2Traj.nbConfig = " << nbConfig);
  ODEBUG2("walkTime.size() = " << walkTime.size());

  return ETHER;
}


/* --- GetConfig ----------------------------------------------------- */

/** walkGetConfigMain  -  codel EXEC of GetConfig
 *
 * Reports:      OK
 *
 * \return EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
walkGetConfigMain(int *report)
{
  double robotConfig[40];
  ghrp2_target_str cfg;
  POSTER_ID p;
  STATUS s;
  ktStatus r;
  int i;

  s = posterFind("hrp2State", &p);
  if (s != OK) { *report = S_walk_NO_ROBOT_STATE; return ETHER; }

  if (posterRead(p, offsetof(ghrp2_robotstate_str, target),
		 &cfg, sizeof(cfg)) != sizeof(cfg)) {
    *report = S_walk_NO_ROBOT_STATE; return ETHER;
  }

  for(i=0; i<40; i++) {
    robotConfig[i] = cfg.angle[i];
  }

  r = hppPlanner->initOpenHRPPosturePatGenRad(robotConfig);
  if (r != KD_OK) {
    *report = S_walk_NO_SOLUTION;
    return ETHER;
  }

  return ETHER;
}


/* --- SavePath ------------------------------------------------------ */

/** walkSavePathMain  -  codel EXEC of SavePath
 *
 * Reports:      OK
 *              S_walk_NO_SUCH_FILE
 *
 * \returns EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
walkSavePathMain(WalkName *name, int *report)
{
  FILE *fpos, *frpy, *fzmp;
  char full[1024];
  ktStatus s;

  vector<CkwsConfigShPtr> path;
  vector<double> walkTime;
  vector< vector<double> > zmp;
  vector< vector<double> > waist;
  unsigned int walkIndex, i;

  /* get data */
  s = hppPlanner->solution(path, walkTime, zmp, waist);
  if (path.size() < 1) return ETHER;

  /* open pos */
  sprintf(full, "%s.pos", name->name);
  fpos = fopen(full, "w");
  if (!fpos) {
    *report = S_walk_NO_SUCH_FILE; return ETHER;
  }

  /* open rpy */
  sprintf(full, "%s.rpy", name->name);
  frpy = fopen(full, "w");
  if (!frpy) {
    fclose(fpos);
    *report = S_walk_NO_SUCH_FILE; return ETHER;
  }

  /* open zmp */
  sprintf(full, "%s.zmp", name->name);
  fzmp = fopen(full, "w");
  if (!fzmp) {
    fclose(fpos); fclose(frpy);
    *report = S_walk_NO_SUCH_FILE; return ETHER;
  }

  const ChppDeviceShPtr hrp2 =
    KIT_DYNAMIC_PTR_CAST(ChppDevice, hppPlanner->robotIthProblem(1));

  if (!hrp2) {
    *report = S_walk_FAILED_TO_GET_HRP2;
    return ETHER;
  }

  std::vector<double> kwsConfigDofVector;
  vectorN jrlDynamicsDofVector(hrp2->numberDof());

  /* write data */
  for (walkIndex = 0; walkIndex < walkTime.size(); walkIndex++) {
    fprintf(fpos, "%g ", walkTime[walkIndex]);
    fprintf(frpy, "%g ", walkTime[walkIndex]);
    fprintf(fzmp, "%g ", walkTime[walkIndex]);

    /*
       Export configuration along path in vector
    */
    if (path[walkIndex]->getDofValues(kwsConfigDofVector) != KD_OK) {
      *report = S_walk_CONFIG_CONVERSION_FAILED;
      return ETHER;
    }

    /*
      Convert Kineo vector of dof values into
      OpenHRP configuration vector
    */
    if (!hrp2->kwsToJrlDynamicsDofValues(kwsConfigDofVector, jrlDynamicsDofVector)) {
      *report = S_walk_CONFIG_CONVERSION_FAILED;
      return ETHER;
    }

    /*
       The first 6 degrees of freedom represent the configuration
       of the waist
    */
    for(i = 6; i<jrlDynamicsDofVector.size(); i++) {
      fprintf(fpos, "%g ", jrlDynamicsDofVector[i]);
    }
    fprintf(fpos, "\n");

    fprintf(frpy, "%g %g %g\n",
	    waist[walkIndex][3], waist[walkIndex][4], waist[walkIndex][5]);

    fprintf(fzmp, "%g %g %g\n",
	    zmp[walkIndex][0], zmp[walkIndex][1], zmp[walkIndex][2]);
  }

  /* done */
  fclose(fpos); fclose(frpy); fclose(fzmp);
  return ETHER;
}


/* --- ExecutePath --------------------------------------------------- */

/** walkExecutePathStart  -  codel START of ExecutePath
 *
 * \return OK
 */

static vector<CkwsConfigShPtr> path;
static vector<double> walkTime;
static vector< vector<double> > zmp;
static vector< vector<double> > waist;
static unsigned int walkIndex;

ACTIVITY_EVENT
walkExecutePathStart(int *report)
{
  ktStatus s;

  if (!SDI_F->initialized) { *report = S_walk_NOT_INITIALIZED; return ETHER; }

  s = hppPlanner->solution(path, walkTime, zmp, waist);
  walkIndex = 0;
  if (path.size() < 1) return ETHER;
  return EXEC;
}

/** walkExecutePathMain  -  codel EXEC of ExecutePath
 *
 * \return OK
 */

ACTIVITY_EVENT
walkExecutePathMain(int *report)
{
  static unsigned int uid;

  int i, n;
  ghrp2_target_str cfg;

  const ChppDeviceShPtr hrp2 =
    KIT_DYNAMIC_PTR_CAST(ChppDevice, hppPlanner->robotIthProblem(1));

  if (!hrp2) {
    *report = S_walk_FAILED_TO_GET_HRP2;
    return ETHER;
  }

  std::vector<double> kwsConfigDofVector;
  vectorN jrlDynamicsDofVector(hrp2->numberDof());

  /* flood fill the buffer */
  ODEBUG2("walkTime.size() = " << walkTime.size());

  for (walkIndex = 0; walkIndex < walkTime.size(); walkIndex++) {
    if (walkIndex > 0) {
      cfg.time = walkTime[walkIndex] - walkTime[walkIndex-1];
      uid++;
    } else {
      cfg.time = 20e-3;
      uid = 0;
    }
    cfg.uid = uid;

    /*
       Export configuration along path in vector
    */
    if (path[walkIndex]->getDofValues(kwsConfigDofVector) != KD_OK) {
      *report = S_walk_CONFIG_CONVERSION_FAILED;
      return ETHER;
    }

    /*
      Convert Kineo vector of dof values into
      OpenHRP configuration vector
    */
    if (!hrp2->kwsToJrlDynamicsDofValues(kwsConfigDofVector, jrlDynamicsDofVector)) {
      *report = S_walk_CONFIG_CONVERSION_FAILED;
      return ETHER;
    }

    /*
       The first 6 degrees of freedom represent the configuration
       of the waist
    */
    for(i = 0; i<GHRP2_MAXDOF; i++) cfg.angle[i] = jrlDynamicsDofVector[i+6];

    cfg.zmp[0] = zmp[walkIndex][0];
    cfg.zmp[1] = zmp[walkIndex][1];
    cfg.zmp[2] = zmp[walkIndex][2];

    {
      double ca, sa, cb, sb, cg, sg;

      ca = cos(waist[walkIndex][5]);	sa = sin(waist[walkIndex][5]);
      cb = cos(waist[walkIndex][4]);	sb = sin(waist[walkIndex][4]);
      cg = cos(waist[walkIndex][3]);	sg = sin(waist[walkIndex][3]);

      cfg.waist_att[0] = ca*cb;
      cfg.waist_att[1] = ca*sb*sg - sa*cg;
      cfg.waist_att[2] = ca*sb*cg + sa*sg;

      cfg.waist_att[3] = sa*cb;
      cfg.waist_att[4] = sa*sb*sg + ca*cg;
      cfg.waist_att[5] = sa*sb*cg - ca*sg;

      cfg.waist_att[6] = -sb;
      cfg.waist_att[7] = cb*sg;
      cfg.waist_att[8] = cb*cg;
    }

    n = netshring_push(shtarget, &cfg, sizeof(cfg));
    if (n < 0 || n != sizeof(cfg)) {
      *report = S_walk_CANNOT_PUSH;
      return ETHER;
    }
  }

  cfg.time = -1.0;
  n = netshring_push(shtarget, &cfg, sizeof(cfg));
  if (n < 0 || n != sizeof(cfg)) {
    *report = S_walk_CANNOT_PUSH;
    return ETHER;
  }

  return ETHER;
}
